/*
 * Copyright (c) 2023–2024 DTB <trinity@trinity.moe>
 * SPDX-License-Identifier: AGPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/.
 */

#include <signal.h> /* sigaction(2), signal(2), struct sigaction, SIGINT */
#include <stdio.h> /* fprintf(3), fgetc(3), perror(3), fputc(3), stderr, stdin,
                    * stdout, EOF, NULL */
#include <stdlib.h> /* exit(3), EXIT_FAILURE */
#if !defined EX_OK || !defined EX_OSERR || !defined EX_USAGE
#	include <sysexits.h>
#endif
#include <termios.h> /* tcgetattr(3), tcsetattr(3), struct termios, ECHO */
#include <unistd.h> /* getopt(3), isatty(3), STDIN_FILENO */

static char *program_name = "peek";

/* Restores terminal echo; otherwise when a user ^Cs the terminal would
 * continue to not display typed text. If sig isn't zero, this will terminate
 * the program. */
static void restore_echo(int sig){
	static struct termios t;

	tcgetattr(STDIN_FILENO, &t);
	t.c_lflag |= ECHO;
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &t);

	/* If, for whatever ungodly reason, exit(3) returns, the user will notice
	 * their typed characters on the screen. */
	if(sig != 0)
		exit(EXIT_FAILURE);

	return;
}

static int oserr(char *s){ perror(s); return EX_OSERR; }
static int ioerr(char *s){ perror(s); restore_echo(0); return EX_IOERR; }
static int usage(char *s){
	fprintf(stderr, "Usage: %s (-1enot) (-p [program [arguments...]])\n", s);
	return EX_USAGE;
}

int main(int argc, char *argv[]){

	if(argc < 1)
		return usage(program_name);

	{ /* options parsing */
		char allow_nonterminals;
		int c;

		allow_nonterminals = 0;
		while((c = getopt(argc, argv, "i")) != -1)
			switch(c){
			case 'i':	allow_nonterminals = 1; break;
			default:	return usage(argv[0]);
			}

		if(argc > optind)
			return usage(argv[0]);

		if(!allow_nonterminals && isatty(STDIN_FILENO) != 1){
			fprintf(stderr, "%s: Must be run in a terminal"
			    " (option -i skips this check)\n", argv[0]);
			return EX_USAGE;
		}
	}

	{ /* install signal handler */
		/* There isn't a difference in functionality between the signal(2) and
		 * sigaction(2) methods. sigaction(2) is vastly preferred for
		 * portability but some configurations can only use signal(2). */
		/* Errors aren't terminating because the worst that happens is some
		 * terminal phooeyness if things go awry. */
#if defined _POSIX_C_SOURCE
		struct sigaction act = { 0 };

		act.sa_handler = restore_echo;
		if(sigaction(SIGINT, &act, NULL) != 0)
			perror(argv[0]);
#else
		if(signal(SIGINT, restore_echo) == SIG_ERR)
			perror(argv[0]);
#endif
	}
	
	{ /* banish terminal echo */
		struct termios t;

		tcgetattr(STDIN_FILENO, &t);
		t.c_lflag ^= ECHO;
		tcsetattr(STDIN_FILENO, TCSAFLUSH, &t);
	}

	{ /* actual input loop */
		int c;

		while((c = getc(stdin)) != EOF)
			if(putc(c, stdout) == EOF)
				return ioerr(argv[0]);
	}

	restore_echo(0);

	return EX_OK;
}
